use crate::{
    allocator,
    ffi_util::{import, Nullable},
    util::data_stream::ChecksumEncoder,
};

use super::{
    avatar::LogicClientAvatar, battle::LogicBattle, command::LogicCommandManager,
    home::LogicClientHome, time::LogicTime,
};

#[repr(C)]
pub struct LogicGameMode {
    unk_gap_1: [u8; 36],                                         // 0-36
    pub logic_command_manager: &'static mut LogicCommandManager, // 36
    game_object_manager: usize,                                  // 40 // TODO!
    achievement_manager: usize,                                  // 44 // TODO!
    pub home: Nullable<LogicClientHome>,                         // 48
    pub player_avatar: Nullable<LogicClientAvatar>,              // 52
    pub time: LogicTime,                                         // 56
    unk_gap_2: [u8; 8],                                          // 116 - 124
    pub battle: Nullable<LogicBattle>,                           // 124
    unk_gap_3: [u8; 168],                                        // 128 - 296
}

impl LogicGameMode {
    pub fn new(is_battle: bool) -> &'static mut Self {
        import!(logic_game_mode_ctor(ptr: *mut LogicGameMode, is_battle: bool) -> () = 0x119DAC + 1);
        let instance = allocator::new::<Self>();
        logic_game_mode_ctor(instance, is_battle);
        unsafe { &mut *instance }
    }

    pub fn load_home_state(
        &mut self,
        logic_client_home: *const LogicClientHome,
        logic_client_avatar: *const LogicClientAvatar,
        home_passed_time_seconds: i32,
        home_passed_time_ticks: i32,
        home_random_seed: i32,
    ) {
        import!(logic_game_mode_load_home_state(lgm: *mut LogicGameMode, lch: *const LogicClientHome, lca: *const LogicClientAvatar, hpts: i32, hptt: i32, hrs: i32) -> () = 0x11A274 + 1);

        logic_game_mode_load_home_state(
            self,
            logic_client_home,
            logic_client_avatar,
            home_passed_time_seconds,
            home_passed_time_ticks,
            home_random_seed,
        );
    }

    pub fn add_player(&mut self, player_avatar: &LogicClientAvatar) {
        import!(logic_game_mode_add_player(lgm: *mut LogicGameMode, player_avatar: *const LogicClientAvatar) -> () = 0x11AD80 + 1);
        logic_game_mode_add_player(self, player_avatar);
    }

    pub fn update_one_tick(&mut self) {
        import!(logic_game_mode_update_one_tick(lgm: *mut LogicGameMode) -> () = 0x11A718 + 1);
        logic_game_mode_update_one_tick(self);
    }

    pub fn calculate_checksum(&self) -> i32 {
        import!(logic_game_mode_calculate_checksum(lgm: *const LogicGameMode) -> i32 = 0x11A85C + 1);
        logic_game_mode_calculate_checksum(self)
    }

    pub fn encode(
        &self,
        encoder: &mut ChecksumEncoder,
        include_commands: bool,
        avatar: &LogicClientAvatar,
        unk: bool,
    ) {
        import!(logic_game_mode_encode(lgm: *const LogicGameMode, encoder: *mut ChecksumEncoder, include_commands: bool, avatar: *const LogicClientAvatar, unk: bool) -> () = 0x11A8D4 + 1);
        logic_game_mode_encode(self, encoder, include_commands, avatar, unk);
    }

    pub fn destruct(&self) {
        import!(logic_game_mode_destruct(lgm: *const LogicGameMode) -> () = 0x11A034 + 1);
        logic_game_mode_destruct(self);
    }
}
