use crate::{
    allocator::malloc,
    array_list::LogicArrayList,
    ffi_util::import,
    logic::data::{LogicDataTables, LogicLocationData},
    sc_string::ScString,
};

#[repr(transparent)]
pub struct DataLoaderFactory(*const u8);

impl DataLoaderFactory {
    pub fn new() -> Self {
        import!(data_loader_factory_ctor(ptr: *const u8) -> () = 0x15EC38 + 1);

        let instance = malloc(4);
        data_loader_factory_ctor(instance);
        Self(instance)
    }
}

struct LogicResources;

impl LogicResources {
    pub fn create_data_table_resources_array() -> &'static LogicArrayList<LogicDataTableResource> {
        import!(logic_resources_create_data_table_resources_array() -> *const u8 = 0x1078E8 + 1);
        unsafe {
            &*(logic_resources_create_data_table_resources_array()
                as *const LogicArrayList<LogicDataTableResource>)
        }
    }

    pub fn load(
        resources_array: &LogicArrayList<LogicDataTableResource>,
        index: usize,
        csv: &CSVNode,
    ) {
        import!(logic_resources_load(r: *const LogicArrayList<LogicDataTableResource>, i: i32, csv: *const u8, a4: i32, a5: i32) -> () = 0x10833C + 1);
        logic_resources_load(
            resources_array as *const LogicArrayList<LogicDataTableResource>,
            index as i32,
            csv.0,
            0,
            0,
        );
    }
}

#[repr(transparent)]
struct LogicDataTableResource(*const u8);

impl LogicDataTableResource {
    pub fn get_file_name(&self) -> String {
        unsafe { (&*(self.0.wrapping_add(8) as *const ScString)).to_string() }
    }
}

#[repr(transparent)]
pub struct CSVNode(pub *const u8);

pub struct ResourceManager;

impl ResourceManager {
    pub fn init(data_loader_factory: &DataLoaderFactory) {
        import!(resource_manager_init(dlf: *const u8, data_path: *const u8) -> () = 0x15D668 + 1);
        resource_manager_init(data_loader_factory.0, [0x00].as_ptr());
    }

    pub fn resource_to_load() -> i32 {
        import!(resource_manager_resource_to_load() -> i32 = 0x15E068 + 1);
        resource_manager_resource_to_load()
    }

    pub fn load_next_resource() {
        import!(resource_manager_load_next_resource() -> () = 0x15DC10 + 1);
        resource_manager_load_next_resource();
    }

    pub fn get_csv(file_name: &str) -> CSVNode {
        import!(resource_manager_get_csv(file_name: *const ScString) -> *const u8 = 0x15E08C + 1);
        CSVNode(resource_manager_get_csv(ScString::new(file_name)))
    }
}

#[repr(transparent)]
struct ResourceListener(*const u8);

impl ResourceListener {
    pub fn new() -> Self {
        import!(resource_listener_ctor(ptr: *const u8) -> () = 0x15C370 + 1);

        let instance = malloc(20);
        resource_listener_ctor(instance);
        Self(instance)
    }

    pub fn add_file(&self, name: &str) {
        import!(resource_listener_add_file(ptr: *const u8, name: *const ScString, a3: i32, a4: i32, a5: i32, a6: i32) -> () = 0x15C51C + 1);
        resource_listener_add_file(self.0, ScString::new(name), -1, -1, -1, -1);
    }

    pub fn start_loading(&self) {
        import!(resource_listener_start_loading(ptr: *const u8) -> () = 0x15C7BC + 1);
        resource_listener_start_loading(self.0);
    }
}

pub fn init() {
    let data_loader_factory = DataLoaderFactory::new();
    ResourceManager::init(&data_loader_factory);

    let listener = ResourceListener::new();
    let data_table_resources_array = LogicResources::create_data_table_resources_array();

    for resource in data_table_resources_array.as_slice().iter() {
        listener.add_file(&resource.get_file_name());
    }

    listener.start_loading();

    while ResourceManager::resource_to_load() != 0 {
        ResourceManager::load_next_resource();
    }

    for (index, resource) in data_table_resources_array.as_slice().iter().enumerate() {
        let csv = ResourceManager::get_csv(&resource.get_file_name());
        LogicResources::load(data_table_resources_array, index, &csv);
    }
}

pub fn prepare_location_data(location: &LogicLocationData) {
    let tilemap_file_name = location.get_tilemap_file_name();
    load_one_resource(&tilemap_file_name);

    let csv = ResourceManager::get_csv(&tilemap_file_name);
    LogicDataTables::init_tilemap(csv, location.parent.global_id % 1_000_000);
}

pub fn load_one_resource(path: &str) {
    import!(game_main_load_asset(path: *const ScString) -> () = 0x5A73C + 1);
    game_main_load_asset(ScString::new(path));
}
