use crate::{
    allocator,
    logic::{
        avatar::LogicClientAvatar,
        data::LogicDataTables,
        home::LogicClientHome,
        json::LogicJSONParser,
        math::LogicLong,
        spell::{LogicSpell, LogicSpellDeck},
    },
    util::data_stream::ByteStream,
};

pub struct PlayerSaveData {
    pub logic_client_home: Vec<u8>,
    pub logic_client_avatar: Vec<u8>,
}

impl PlayerSaveData {
    pub fn new(
        logic_client_home: &LogicClientHome,
        logic_client_avatar: &LogicClientAvatar,
    ) -> Self {
        let mut stream = ByteStream::new(64);
        logic_client_home.encode(&mut stream.parent);
        let logic_client_home = stream.get_data().to_vec();
        stream.destruct();

        let mut stream = ByteStream::new(64);
        (logic_client_avatar.vtable.encode)(logic_client_avatar, &mut stream.parent);
        let logic_client_avatar = stream.get_data().to_vec();
        stream.destruct();

        Self {
            logic_client_home,
            logic_client_avatar,
        }
    }

    pub fn create_default_data(id: &LogicLong) -> Self {
        const STARTING_DECK_JSON: &str = r#"[{"d":26000000,"t":0,"s":0,"l":12,"c":1},{"d":26000001,"t":0,"s":0,"l":12,"c":1},{"d":26000013,"t":0,"s":0,"l":12,"c":1},{"d":28000001,"t":0,"s":0,"l":12,"c":1},{"d":28000000,"t":0,"s":0,"l":10,"c":1},{"d":26000003,"t":0,"s":0,"l":10,"c":1},{"d":26000004,"t":0,"s":0,"l":7,"c":1},{"d":26000005,"t":0,"s":0,"l":12,"c":1}]"#;

        let logic_client_home = LogicClientHome::new();
        logic_client_home
            .id
            .set(LogicLong::new(id.higher_int, id.lower_int));

        let starting_deck = LogicJSONParser::parse(STARTING_DECK_JSON).unwrap();
        logic_client_home.spell_deck.load(starting_deck);

        for i in 0..LogicDataTables::get_spell_count() {
            let spell_data = LogicDataTables::get_spell_at(i);
            if !logic_client_home.has_spell(spell_data) {
                let spell = LogicSpell::new();
                spell.data = spell_data;
                spell.level_index = spell_data.get_max_level_index();
                logic_client_home.spell_collection.add_spell(spell);
            }
        }

        let logic_client_avatar = LogicClientAvatar::new();
        logic_client_avatar.id.set(id.higher_int, id.lower_int);
        logic_client_avatar
            .account_id
            .set(id.higher_int, id.lower_int);
        logic_client_avatar.home_id.set(id.higher_int, id.lower_int);
        logic_client_avatar.name.assign("xeondev");
        logic_client_avatar.name_change_state = 0;
        logic_client_avatar.name_set_by_user = true;
        logic_client_avatar.arena = LogicDataTables::get_data_by_id(54000007).unwrap();
        logic_client_avatar.npc_win_count = 8;
        logic_client_avatar.score = 3000;
        logic_client_avatar.level = 13;

        let instance = Self::new(logic_client_home, logic_client_avatar);

        logic_client_home.destruct();
        allocator::free(logic_client_home as *const _);

        (logic_client_avatar.vtable.destruct)(logic_client_avatar);
        allocator::free(logic_client_avatar as *const _);

        instance
    }

    pub fn get_logic_client_home(&self) -> &'static mut LogicClientHome {
        let logic_client_home = LogicClientHome::new();
        let mut stream = ByteStream::from(&self.logic_client_home);
        logic_client_home.decode(&mut stream);
        stream.destruct();

        logic_client_home
    }

    pub fn get_logic_client_avatar(&self) -> &'static mut LogicClientAvatar {
        let logic_client_avatar = LogicClientAvatar::new();
        let mut stream = ByteStream::from(&self.logic_client_avatar);
        (logic_client_avatar.vtable.decode)(logic_client_avatar, &mut stream);
        stream.destruct();

        logic_client_avatar
    }

    pub fn get_spell_deck(&self) -> &'static mut LogicSpellDeck {
        let logic_client_home = self.get_logic_client_home();
        let cloned_deck = logic_client_home.spell_deck.clone();
        logic_client_home.destruct();
        allocator::free(logic_client_home);

        cloned_deck
    }
}
